MODULE  STACKMOD

  IMPLICIT NONE

  PRIVATE
 ! export list for module STACKMOD:                                 !
  PUBLIC :: STACK, INIT, EMPTY, PUSH, POP, TOP, WRITE, DELETE

  TYPE  STACK
    PRIVATE
    TYPE(STACKELEM), POINTER :: TOP
  END TYPE  STACK

  TYPE, PRIVATE :: STACKELEM            ! element type not exported !
    INTEGER :: DATA = 0
    TYPE(STACKELEM), POINTER :: SUCC => NULL()
  END TYPE  STACKELEM

 CONTAINS

 ! initializes stack to a defined state (empty) !
  SUBROUTINE  INIT (S)
    TYPE(STACK), INTENT(OUT) :: S
    NULLIFY (S%TOP)
  END SUBROUTINE  INIT

 ! tests if stack is currently empty !
  FUNCTION  EMPTY (S)
    TYPE(STACK), INTENT(IN) :: S
    LOGICAL                 :: EMPTY
    EMPTY =  .NOT. ASSOCIATED(S%TOP)
  END FUNCTION  EMPTY

 ! adds new top element to stack !
  SUBROUTINE  PUSH (S, ELEM)
    TYPE(STACK), INTENT(INOUT) :: S
    INTEGER, INTENT(IN)  :: ELEM
    TYPE(STACKELEM), POINTER   :: PTR
    INTEGER                    :: err_code
    ALLOCATE (PTR, STAT = err_code)
    IF ( err_code == 0 ) THEN
      PTR%DATA = ELEM
      PTR%SUCC => S%TOP
      S%TOP => PTR
    ELSE
      WRITE(*,*) ' Allocation failed, probably out of memory!'
    END IF
  END SUBROUTINE  PUSH

 ! removes top element from stack and returns its CONTENT !            ! 4 !
 ! via the second argument ELEM, if PRESENT               !
  SUBROUTINE  POP (S, ELEM)
    TYPE(STACK), INTENT(INOUT)           :: S
    INTEGER, INTENT(OUT), OPTIONAL :: ELEM
    TYPE(STACKELEM), POINTER             :: PTR
    INTEGER                              :: err_code
    IF ( .NOT. EMPTY(S) ) THEN
      PTR => S%TOP
      IF ( PRESENT(ELEM) )  ELEM = PTR%DATA
      S%TOP => PTR%SUCC
      DEALLOCATE (PTR, STAT = err_code)
      IF ( err_code /= 0 )                                     &
     &  WRITE(*,*) ' Deallocation failed, inconsistent memory!'
    ELSE
      WRITE(*,*) ' Cannot pop from empty stack!'
    END IF
  END SUBROUTINE  POP

 ! returns CONTENT of top element !
  FUNCTION  TOP (S)
    TYPE(STACK), INTENT(IN) :: S
    INTEGER           :: TOP
    IF ( .NOT. EMPTY(S) ) THEN
      TOP = S%TOP%DATA
    ELSE                !!! Output in a function can be a PROBLEM !!!
      WRITE(*,*) ' Cannot get top of empty stack!'
    END IF
  END FUNCTION  TOP

 ! writes complete stack from top to bottom       !                    ! 5 !
 ! (low level - should be inside module STACKMOD) !
  SUBROUTINE  WRITE (S)
    TYPE(STACK), INTENT(IN)  :: S
    TYPE(STACKELEM), POINTER :: PTR
    PTR => S%TOP
    DO WHILE ( ASSOCIATED(PTR) )
      WRITE(*,FMT="(I3)",ADVANCE="NO") PTR%DATA
      PTR => PTR%SUCC
    END DO
    WRITE(*,*)
  END SUBROUTINE  WRITE

 ! deletes complete stack                         !
 ! (low level - should be inside module STACKMOD) !
  SUBROUTINE  DELETE (S)
    TYPE(STACK), INTENT(INOUT) :: S
    TYPE(STACKELEM), POINTER   :: PTR
    INTEGER                    :: err_code
    DO WHILE ( ASSOCIATED(S%TOP) )
      PTR => S%TOP
      S%TOP => PTR%SUCC
      DEALLOCATE (PTR, STAT = err_code)
      IF ( err_code /= 0 ) THEN
        WRITE(*,*) ' Deallocation failed, inconsistent memory!'
        EXIT   ! >----------------------------------->>> EXIT !
      END IF
    END DO
  END SUBROUTINE  DELETE

END MODULE  STACKMOD



PROGRAM ackermann
  USE STACKMOD
  IMPLICIT NONE

  INTEGER:: i,j,ergeb=0
  REAL:: t1,t2

  DO
    WRITE(*,*) 'Enter integer i and j for computing ackermann_fun(i,j)'
    READ(*,*) i,j
    IF(i<0 .OR. j<0 )EXIT


    CALL CPU_TIME(t1)
    ergeb = ackermann_fun(i,j)
    CALL CPU_TIME(t2)
    WRITE(*,*) 'recursive computtion, result:', ergeb
    WRITE(*,*) 'time to calculate:', t2-t1


    CALL CPU_TIME(t1)
    ergeb = ackermannstack(i,j)
    CALL CPU_TIME(t2)
    WRITE(*,*) 'iterative computation, result:', ergeb
    WRITE(*,*) 'time to calculate:', t2-t1
  END DO

CONTAINS

  RECURSIVE FUNCTION ackermann_fun(m,n) RESULT(res)
    INTEGER,INTENT(IN):: m,n
    INTEGER:: res

    !!! Rekursionsabbruch
    IF(m==0)THEN !!! .AND. n >= 0
      res = n+1
    ELSE IF(n==0)THEN !!!.AND. m >= 1
      res = ackermann_fun(m-1,1) !!! rekursiver Aufruf
    ELSE !IF (m,n >= 1)
      res = ackermann_fun(m-1, ackermann_fun(m,n-1)) !!! zwei rekursive Aufrufe
    END IF
  END FUNCTION ackermann_fun

!!! Jeder rekursiver Algorithmus kann mit Hilfe eines Stacks
!!! in einen iterativen Algorithmus umgewandelt werden.

!!! Die Komplexitaet des Algorithmus ändert sich durch die Umwandlung nicht.
  FUNCTION ackermannstack(x,y)
    INTEGER :: ackermannstack,x,y,p,q,o
    TYPE(STACK)::s

    !!! Im Python-Bsp "a", hier "-1" - als Symbol für Funktionsaufruf

    !!! Initialisieren mit Ursprungsproblem
    CALL INIT(s)
    CALL PUSH(s,-1)
    CALL PUSH(s,x)
    CALL PUSH(s,y)

    !!! Schleife - 'iteraitiv'
    DO
      CALL WRITE(s)
      CALL POP(s,q)
      CALL POP(s,p)
      CALL POP(s,o)

      !!! Fallunterscheidungen wir in der rekursiven Definition
      IF (o==-1)THEN

        IF (p==0) THEN !!! 1. Fall
          CALL PUSH(s,q+1)
        ELSE IF (q==0)THEN !!! 2. Fall
          CALL PUSH(s,-1) !!! Pseudo-rekursiver Aufruf
          CALL PUSH(s,p-1)
          CALL PUSH(s,1)
        ELSE  !!! 3. Fall
          CALL PUSH(s,-1) !!! Pseudo-rekursiver Aufruf
          CALL PUSH(s,p-1)
          CALL PUSH(s,-1) !!! Pseudo-rekursiver Aufruf
          CALL PUSH(s,p)
          CALL PUSH(s,q-1)
        END IF
      END iF

      !!! Wenn nur noch ein Element auf dem Stack ist, dann ist dieses das Ergebnis
      CALL POP(s,ackermannstack)
      IF (EMPTY(s)) EXIT
      !!! Sonst muss das Element zurueck gelegt werden
      CALL PUSH(s,ackermannstack)
    END DO

    CALL WRITE(s)
    !!! Stack wieder loeschen!
    CALL DELETE(s)
  END FUNCTION ackermannstack

END PROGRAM ackermann
